#lang racket/base
;; crossfire: distributed brute force infrastructure
;;
;; Copyright (C) 2020 haskal
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU Affero General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU Affero General Public License for more details.
;;
;; You should have received a copy of the GNU Affero General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>.

(require (only-in data/integer-set integer-set-contents)
         racket/match racket/runtime-path racket/vector
         scribble/text (rename-in scribble/text/output [output scribble-output])
         "manifest.rkt" "pattern.rkt")

(provide generate-support-code generate-manifest)

;; templating infrastructure

(define (eval-template template-path vars)
  (define out (open-output-string))
  (define cs (current-namespace))
  (define output-exp
    (parameterize ([current-namespace (make-base-namespace)])
      (for ([mod (in-list '(scribble/text racket/match))])
        (namespace-attach-module cs mod)
        (namespace-require mod))
      (hash-for-each vars namespace-set-variable-value!)
      (eval `(include/text (file ,(path->string template-path))))))
  (scribble-output output-exp out)
  (get-output-string out))

;; there used to be a racket implementation of the input generator but it majorly violated the
;; design recipe and got obsoleted by the template C version (which is intended to be more portable
;; and faster)

;; ok gamer move time
(define-runtime-path codegen-template "templates/codegen.rktc")
(define-runtime-path header-template "templates/header.rktc")
(define-runtime-path makefile-template "templates/Makefile.rktmake")
(define-runtime-path manifest-template "templates/manifest.rktrktd")

;; TODO : make configurable
(define c-type "uint64_t")
(define c-type-fmt "%lx")

(define (generate-support-code mf)
  (define vars (hash 'pattern (vector-map integer-set-contents (manifest-pattern mf))
                     'mode (first (manifest-data-ref mf 'mode))
                     'c-type c-type
                     'c-type-fmt c-type-fmt))
  (hash
    "crossfire.c" (eval-template codegen-template vars)
    "crossfire.h" (eval-template header-template vars)
    "Makefile" (eval-template makefile-template vars)))

(define (generate-manifest mode)
  (eval-template
    manifest-template
    (hash 'mode mode)))
